/* $Id:$
***************************************************************************

   X server for LibGGI - Handling of input devices

   Copyright (C) 1997 Jason McMullan	[jmcc@cs.cmu.edu]
   Copyright (C) 1997 Michael Krause	[rawstyle@ms.demo.org]
   Copyright (C) 1998 Marcus Sundberg	[marcus@ggi-project.org]

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************
*/

#include "X11/X.h"
#define NEED_EVENTS
#include "X11/Xproto.h"

#include "scrnintstr.h"
#include "inputstr.h"
#include "mipointer.h"
#include "input.h"
#include "mi.h"

#include "xggi.h"

#define MAXBUTTONS 32

static int
MouseAccelerate(DeviceIntPtr device, int delta)
{
	int  sgn = sign(delta);
	PtrCtrl *pCtrl;
	short ret;

	delta = abs(delta);
	pCtrl = &device->ptrfeed->ctrl;
	if (delta > pCtrl->threshold) {
		ret = (short) sgn *
			(pCtrl->threshold + ((delta - pCtrl->threshold)
					     * pCtrl->num) / pCtrl->den);
	} else {
		ret = (short) sgn * delta;
	}
	return ret;
}

static void
xggiQueueMouseEvent(ggi_event *ev)
{
	DevicePtr pMouse;
	xEvent xev;
	int dx, dy;
	static int oldx = 0, oldy = 0;
    
	pMouse = LookupPointerDevice();
	if (!pMouse->on)
		return;
    
	xev.u.keyButtonPointer.time = TVAL2TIME(ev->any.time);
    
	switch (ev->any.type) {
	case evPtrAbsolute:
		dx = MouseAccelerate((DeviceIntPtr)pMouse, ev->pmove.x - oldx);
		dy = MouseAccelerate((DeviceIntPtr)pMouse, ev->pmove.y - oldy);
		oldx = ev->pmove.x;
		oldy = ev->pmove.y;
		miPointerDeltaCursor(dx, dy, xev.u.keyButtonPointer.time);
		break;

	case evPtrRelative:
		dx = MouseAccelerate((DeviceIntPtr)pMouse, ev->pmove.x);
		dy = MouseAccelerate((DeviceIntPtr)pMouse, ev->pmove.y);
		oldx += ev->pmove.x;
		oldy += ev->pmove.y;
		miPointerDeltaCursor(dx, dy, xev.u.keyButtonPointer.time);
		break; 

	case evPtrButtonPress:
		if (ev->pbutton.button & 2) {
			/* Right button */
			xev.u.u.detail = 3; 
		} else if (ev->pbutton.button & 4) {
			/* Middle button */
			xev.u.u.detail = 2;
		} else {
			/* Other button */
			uint32 i, val = 1;
			for (i = 1; i <= MAXBUTTONS; i++) {
				if (ev->pbutton.button & val) {
					xev.u.u.detail = i;
					break;
				}
				val <<= 1;
			}
		}
		xev.u.u.type = ButtonPress;
		mieqEnqueue(&xev);
		break;

	case evPtrButtonRelease:
		if (ev->pbutton.button & 2) {
			/* Right button */
			xev.u.u.detail = 3; 
		} else if (ev->pbutton.button & 4) {
			/* Middle button */
			xev.u.u.detail = 2;
		} else {
			/* Other button */
			uint32 i, val = 1;
			for (i = 1; i <= MAXBUTTONS; i++) {
				if (ev->pbutton.button & val) {
					xev.u.u.detail = i;
					break;
				}
				val <<= 1;
			}
		}
		xev.u.u.type = ButtonRelease;
		mieqEnqueue(&xev);
		break;
	}

	return;
}
  
static int
xggiMouseProc(DeviceIntPtr pDevice, int onoff)
{
	BYTE map[MAXBUTTONS+1];
	int i;
	DevicePtr pDev = (DevicePtr)pDevice;

	switch (onoff) {
	case DEVICE_INIT:
		for (i = 0; i <= MAXBUTTONS; i++) {
			map[i] = i;
		}
		ggiSetEventMask(MASTER_VIS, emKeyboard | emPointer);
		InitPointerDeviceStruct(pDev, map, MAXBUTTONS,
					miPointerGetMotionEvents,
					(PtrCtrlProcPtr)NoopDDA,
					miPointerGetMotionBufferSize());
		break;
	
	case DEVICE_ON:
		pDev->on = TRUE;
		break;
	
	case DEVICE_OFF:
		pDev->on = FALSE;
		break;
	
	case DEVICE_CLOSE:
		break;
	}

	return Success;
}

void
ProcessInputEvents()
{
	mieqProcessInputEvents();
	miPointerUpdate();
}

void
AbortDDX()
{
	ggiExit();
}

static void
xggiDispatchEvents()
{
	ggi_event ev;
	struct timeval ztime = {0,0};
    
	while (ggiEventPoll(MASTER_VIS, emAll, &ztime)) {
		ggiEventRead(MASTER_VIS, &ev, emAll);
		
		switch (ev.any.type) {
		case evPtrRelative:
		case evPtrAbsolute:
		case evPtrButtonPress:
		case evPtrButtonRelease:
			xggiQueueMouseEvent(&ev);
			break;
		case evKeyPress:
		case evKeyRepeat:
		case evKeyRelease:
			xggiQueueKeyboardEvent(&ev);
			break;
		default:
			break;
		}
		ztime.tv_sec = 0;
		ztime.tv_usec = 0;
	}
}

static CARD32
xggiSnarf(OsTimerPtr timer, CARD32 time, pointer data)
{
	xggiDispatchEvents();
	return 20 /* Millis */;
}

static void
xggiInputTimer()
{
	static OsTimerPtr timer;
	timer = TimerSet(NULL, 0, 20 /* Millis */, xggiSnarf, NULL);
}

/*****************************
 * Initailize input handlers *
 *****************************/
void
InitInput(int argc, char *argv[])
{
	DeviceIntPtr p, k;

	p = AddInputDevice(xggiMouseProc, TRUE);
	k = AddInputDevice(xggiKeybdProc, TRUE);
	RegisterPointerDevice(p);
	RegisterKeyboardDevice(k);
	miRegisterPointerDevice(screenInfo.screens[0], p);
	mieqInit((DevicePtr) k, (DevicePtr) p);

	xggiInputTimer();
}
